package pong.view;

import javafx.animation.AnimationTimer;
import javafx.application.Application;
import javafx.event.ActionEvent;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.CheckMenuItem;
import javafx.scene.control.Menu;
import javafx.scene.control.MenuBar;
import javafx.scene.control.MenuItem;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import pong.model.*;
import pong.service.EventService;
import pong.service.ImageService;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static java.lang.System.out;
import static pong.model.Pong.*;
import static pong.service.EventService.PongEvent;
import static pong.service.EventService.Type;

/*
 * A JavaFX GUI for the Pong game.
 * No application logic here just GUI and event handling
 *
 * Run this to run the game
 *
 * See: https://en.wikipedia.org/wiki/Pong
 *
 */
public class PongGUI extends Application {

    // The game
    private Pong pong;
    private boolean running = false;    // Is game running?

    // ------- Keyboard handling ----------------------------------

    private final int speed = 3;

    private void keyPressed(KeyEvent event) {
        if (!running) {
            return;
        }
        KeyCode kc = event.getCode();
        switch (kc) {
            case UP:
                IMovable p = pong.getRightPaddle();
                p.setDy(-speed);
                break;
            case DOWN:
                p = pong.getRightPaddle();
                p.setDy(speed);
                break;
            case Q:
                p = pong.getLeftPaddle();
                p.setDy(-speed);
                break;
            case A:
                p = pong.getLeftPaddle();
                p.setDy(speed);
                break;
            default:  // Nothing
        }
    }

    private void keyReleased(KeyEvent event) {
        if (!running) {
            return;
        }
        KeyCode kc = event.getCode();
        switch (kc) {
            case UP:
            case DOWN:
                IMovable p = pong.getRightPaddle();
                p.setDy(0);
                break;
            case A:
            case Q:
                p = pong.getLeftPaddle();
                p.setDy(0);
                break;
            default: // Nothing
        }
    }

    // --- Handling events coming form the model -----

    private void handlePongEvent(PongEvent evt) {
        switch (evt.type) {
            case HIT_PADDLE:
                Theme.playSound(Type.HIT_PADDLE);
                break;
            case HIT_WALL:
                //Theme.playSound(...);
                break;
            case BALL_OUT:
                //Theme.playSound(...);
                break;
        }
    }

    // ---------- Menu actions ---------------------

    private void newGame() {
        Theme.setTheme(themeName);
        fixMenuForNewGame();

        // Build the model
        Paddle pRight = new Paddle(WIDTH - 50, HEIGHT / 2 + PADDLE_HEIGHT / 2, PADDLE_WIDTH, PADDLE_HEIGHT, "blue");
        Paddle pLeft = new Paddle(50, HEIGHT / 2 + PADDLE_HEIGHT / 2, PADDLE_WIDTH, PADDLE_HEIGHT, "red");
        List<Wall> walls = new ArrayList<>(Arrays.asList(
                new Wall(0, 0, 3, HEIGHT),
                new Wall(WIDTH - 3, 0, 3, HEIGHT)));
        List<Floor> floors = new ArrayList<>(Arrays.asList(
                new Floor(10, 0, WIDTH - 20, 3),
                new Floor(10, HEIGHT - 3, WIDTH - 20, 3)));
        pong = new Pong(pLeft, pRight, new Ball(), walls, floors);

        // Start game
        timer.start();
        Theme.playMusic(0.5);
        running = true;
    }

    private void killRound() {
        // Stop game
        timer.stop();
        running = false;
        Theme.stopMusic();
        fixMenuForKillRound();
    }


    // *********************  Nothing to do below this  ***********************

    // ---- Menu handling -----------------

    private void handleMenu(ActionEvent e) {
        MenuItem mi = (MenuItem) e.getSource();
        out.println("Click " + mi.getText());
        String menuText = mi.getText();
        switch (menuText) {
            case "New":
                newGame();
                break;
            case "Stop":
                killRound();
                break;
            case "Exit":
                System.exit(0);
                break;
            case "Cool":
            case "Duckie":
                setTheme(menuText);
                break;
            case "Play":
                toggleMusic();
                break;
            default: // Nothing
        }
    }


    private Theme.Name themeName = Theme.Name.COOL;   // COOL is default theme

    private void setTheme(String name) {
        unSelectAllCheckedMenuItems("Themes");
        setCheckedMenuItemSelected(name, true);
        switch (name) {
            case "Cool":
                themeName = Theme.Name.COOL;
                break;
            case "Duckie":
                themeName = Theme.Name.DUCKIE;
                break;
        }
    }

    private void toggleMusic() {
        if (Theme.isPlayingMusic()) {
            Theme.stopMusic();
        } else {
            Theme.playMusic(0.5);
        }
    }

    private MenuBar menuBar;  // Created in start()

    private void fixMenuForNewGame() {
        setMenuItemDisable("New", true);
        setMenuItemDisable("Stop", false);
        setMenuDisable("Themes", true);
        setMenuDisable("Music", false);
        setCheckedMenuItemSelected("Play", true);
    }

    private void fixMenuForKillRound() {
        setMenuItemDisable("New", false);
        setMenuItemDisable("Stop", true);
        setMenuDisable("Themes", false);
        setMenuDisable("Music", true);
    }

    private void setMenuDisable(String menuName, boolean b) {
        menuBar.getMenus().filtered(i -> i.getText().equals(menuName)).get(0).setDisable(b);
    }

    private void setMenuItemDisable(String menuItemName, boolean b) {
        menuBar.getMenus().forEach(m -> {
            List<MenuItem> is = m.getItems().filtered(i -> i.getText().equals(menuItemName));
            if (is.size() > 0) {
                is.get(0).setDisable(b);
            }
        });
    }

    private void setCheckedMenuItemSelected(String name, boolean b) {
        menuBar.getMenus().forEach(m -> {
            List<MenuItem> is = m.getItems().filtered(i -> i.getText().equals(name));
            if (is.size() > 0) {
                ((CheckMenuItem) is.get(0)).setSelected(b);
            }
        });
    }

    private void unSelectAllCheckedMenuItems(String menuName) {
        menuBar.getMenus()
                .filtered(i -> i.getText().equals(menuName))
                .get(0)
                .getItems()
                .forEach(mi -> ((CheckMenuItem) mi).setSelected(false));
    }

    // ---------- Rendering -----------------
    // Don't need to understand in depth
    private AnimationTimer timer;

    // For debugging, see renderGame
    private boolean renderDebug = false; //true;

    private void renderGame(GraphicsContext g) {
        g.clearRect(0, 0, WIDTH, HEIGHT);    // Clear everything
        List<IDrawable> drawables = pong.getDrawables();


        if (renderDebug) {                // <---------------- Debugging
            renderDebug(g, drawables);
        } else {
            g.drawImage(Theme.getBackground(), 0, 0); // Better to render else where?!
            g.setFill(Theme.getTextColor());
            g.fillText("Points: " + pong.getPointsLeft(), 10, 20);
            g.fillText("Points: " + pong.getPointsRight(), 500, 20);
            render(g, drawables);
        }
    }

    private void render(GraphicsContext g, List<IDrawable> drawables) {
        for (IDrawable d : drawables) {
            Image i = Theme.getImage(d);
            if( i != null) {
                g.drawImage(i, d.getMinX(), d.getMinY(), d.getWidth(), d.getHeight());
            }
        }
    }


    private void renderDebug(GraphicsContext g, List<IDrawable> drawables) {
        for (IDrawable ds : drawables) {
            g.strokeRect(ds.getMinX(), ds.getMinY(),
                    ds.getWidth(), ds.getHeight());
        }
    }

    @Override
    public void start(Stage primaryStage) throws Exception {

        BorderPane root = new BorderPane();

        if (!renderDebug) {
            Image pongBg = ImageService.getImage("../assets/img/", "pong.jpg", WIDTH, HEIGHT, false);
            root.getChildren().addAll(new ImageView(pongBg));
        }
        // Menu
        menuBar = getMenuBar();
        menuBar.prefWidthProperty().bind(primaryStage.widthProperty());
        root.setTop(menuBar);

        // Default values fo menus
        setCheckedMenuItemSelected("Cool", true);
        setCheckedMenuItemSelected("Play", true);
        setMenuDisable("Music", true);
        setMenuItemDisable("New", false);
        setMenuItemDisable("Stop", true);

        //Group root = new Group();
        Canvas canvas = new Canvas(WIDTH, HEIGHT);
        root.setCenter(canvas);
        GraphicsContext gc = canvas.getGraphicsContext2D();

        timer = new AnimationTimer() {
            // Param not used
            public void handle(long currentNanoTime) {
                pong.update();
                PongGUI.this.handlePongEvent(EventService.remove());
                renderGame(gc);

            }
        };

        Scene scene = new Scene(root);
        scene.setOnKeyPressed(this::keyPressed);
        scene.setOnKeyReleased(this::keyReleased);
        primaryStage.setScene(scene);
        primaryStage.setTitle("Pong");
        primaryStage.show();
    }

    // ----------- Menu -------------------
    // Don't need to understand in depth
    private MenuBar getMenuBar() {
        Menu menuFile = new Menu("File");
        MenuItem[] menuFileItems = {
                new MenuItem("New"),
                new MenuItem("Stop"),
                new MenuItem("Exit"),
        };
        menuFile.getItems().addAll(menuFileItems);

        Menu menuTheme = new Menu("Themes");
        MenuItem[] menuThemeItems = new MenuItem[]{
                new CheckMenuItem("Cool"),
                new CheckMenuItem("Duckie"),
        };
        menuTheme.getItems().addAll(menuThemeItems);

        Menu menuMusic = new Menu("Music");
        MenuItem[] menuMusicItems = new MenuItem[]{
                new CheckMenuItem("Play")
        };
        menuMusic.getItems().addAll(menuMusicItems);

        // -- The bar for all menus
        MenuBar menuBar = new MenuBar();
        menuBar.getMenus().addAll(menuFile, menuTheme, menuMusic);

        // Set same event handler for all
        menuBar.getMenus().forEach(m -> {
            m.getItems().forEach(mi -> {
                mi.setOnAction(this::handleMenu);
            });
        });
        return menuBar;
    }

    public static void main(String[] args) {
        launch(args);
    }
}
